namespace eval VELaSSCo {
    variable _dir
    variable _images
    variable _tunnel_used 0
    variable _sessionID   ""
    variable _modelID     ""
    variable _modelInfo
    variable _status_wid  ""
    variable _login_icon_wid     ""
    variable _icon_do_connection connect_creating.png
    variable _icon_connected     connect_established.png
    variable _cache
    variable _selected
    variable _busy
    variable _ogl
    # maximum vervosity
    variable _verbose_level      9
    # verbose levels:
    # 0 - Nothing
    # 1 - show queries issued with short response

    # utils, Ini & End
    proc GetFullImageName { filename} {
	variable _dir
	return [ file join $_dir images $filename]
    }
    proc GetImage { filename} {
	variable _dir
	variable _images
	if { ![ info exist _images($filename)]} {
	    set _images($filename) [ image create photo -file \
					 [ file join $_dir images $filename]]
	} 
	return $_images($filename)
    }

    proc Ini { dir} {
	variable _dir

	set _dir $dir

	set VersionRequired "12.0.0"
	set comp -1
	catch { 
	    set comp [ GiDVersionCmp $VersionRequired]
	}

	# doing package require ...
	if { [ string first $_dir $::auto_path] == -1} {
	    lappend ::auto_path ${_dir}
	}

	set err_txt ""
	set err [ catch {
	    set ret [ package require gid_velassco]
	    # W "package require gid_velassco = $ret"
	} err_txt ]
	if { $err} {
	    # problems loading the package ( i.e. library)
	    W $err_txt
	    if { $::tcl_platform(platform) == "windows"} {
		set suggestion "\nThe package has been compiled with\n"
		append suggestion "Microsoft Visual C++ 2012\n"
		append suggestion "May be the solution is to install the\n"
		append suggestion "Microsoft Cisual C++ 2012 Redistributable (x64)\n"
		append suggestion "which can be found at [ file join $_dir bin vcredist_x64.exe]"
		W $suggestion
		WarnWin $suggestion
	    } else {
		W "Dependencies:"
		set lib_name [ file join $_dir gid_velassco [ Get_Library_gid_velassco_name gid_velassco]]
		W [ exec ldd $lib_name]
		set suggestion "\nMay be the solution is to install\n"
		append suggestion "boost >= 1.58\n"
		append suggestion "thrift >= 0.92\n"
		append suggestion "openssl >= 1.0.1p\n"
		append suggestion "libevent >= 2.0"
		W $suggestion
	    }
	}

	return
    }

    proc End { args } {
	DoLogout
    }

    proc WaitState { } {
	::GidUtils::WaitState .gid
    }
    
    proc EndWaitState { } {
	::GidUtils::EndWaitState .gid
    }

    proc Print { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
    }

    proc Verbose { level msg} {
	variable _verbose_level
	if { $level < $_verbose_level} {
	    Print $msg
	}
    }

    proc Debug { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
    }

    # functionalities / Vqueries
    proc About { } {
	set package_version "N/A"
	catch {
	    set package_version [ package require gid_velassco]
	}
	# InfoWin "About VELaSSCo:\n\nUsing gid_velassco v. $package_version\nwith support for\n[ gid_velassco help]\n\nPlease visit http://www.velassco.eu"

	set w .gid.wVelasscoAbout
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "About VELaSSCo..."]
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Ok"]] \
	    [ list [ list destroy $w]] \
	    [ list <Return>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "About VELaSSCo"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.about
	ttk::label $w.top.about.l0 -text "Using gid_velassco v. $package_version"
	ttk::label $w.top.about.l1 -text "with support for:"
        ttk::scrollbar $w.top.about.scrolly -command "$w.top.about.t yview" -orient vertical
	text $w.top.about.t -wrap word -width 20 \
	    -yscrollcommand "$w.top.about.scrolly set" \
	    -border 0 -background [ $w cget -background]
	# -height 8
	VisitWebButton $w.top.about.b http://www.velassco.eu "Please visit www.velassco.eu"

	grid $w.top.about.l0 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.l1 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.t -sticky news -padx 4 -pady 0
	grid $w.top.about.scrolly -row 2 -column 1 -sticky ns
	grid $w.top.about.b -sticky ew -padx 4 -pady 0 -columnspan 2
	grid columnconfigure $w.top.about 0 -weight 1
	grid rowconfigure $w.top.about 2 -weight 1
	# grid rowconfigure $w.top.about 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.about -sticky ew -padx 4 -pady 4
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# get rid of gid_velassco ... ?parameters?
	regsub -all { \| } [ lrange [ gid_velassco help] 1 end-1] \n txt
	$w.top.about.t insert end $txt
	# $w.top.about.t insert end ""
	# $w.top.about.t see end
	$w.top.about.t configure -state disabled
    }

    proc OpenModel { { w .gid.wVelasscoOpenModel}} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLogin]} { return}
	# Debug "VQuery OpenModel( $::VELaSSCo(SelectedModel))"
	if { "$::VELaSSCo(SelectedModel)" != ""} {
	    # should verify if there is a model already opened
	    if { "$_modelID" != ""} {
		CloseModel
	    }
	    set err_txt "Ok"
	    set err [ catch {
		set _modelID [ gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw]
	    } err_txt]
	    Verbose 1 "gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw = $_modelID (error = err - $err_txt)"
	    if { !$err} {
		wm title .gid $::VELaSSCo(SelectedModel)
		set _modelInfo(unique_name) $::VELaSSCo(SelectedModel)
	    } else {
		# clean any Postprocess information inside GiD:
		GiD_Process Mescape Files New Yes escape
	    }
	}
    }

    proc CloseModel {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo
	variable _cache
	variable _selected
	variable _busy
	if { ![ CheckLogin]} { return}
	if { "$_modelID" != ""} {
	    set err_txt "Ok"
	    set err [ catch {
		gid_velassco CloseModel $_sessionID $_modelID
	    } err_txt]
	    Verbose 1 "gid_velassco CloseModel $_sessionID $_modelID = $err - $err_txt"
	    set _modelID ""
	    OGLunregisterAndDeleteLists
	    GiD_Process Mescape Files New Yes escape
	    catch {
		array unset _modelInfo
		array unset _cache
		array unset _selected
		array unset _busy
	    }
	}
    }

    proc DoSelectModel { w} {
	variable _sessionID
	variable _modelID
	
	if { $::VELaSSCo(SelectedModel) != ""} {
	    destroy $w
	    OpenModel
	    Verbose 1 "selected model = modelID = $_modelID"
	} else {
	    ErrorWin [_ "No model has been selected"]
	}
    }

    proc DoUpdateThumbImage { wThumb img} {
	if { [ winfo exists $wThumb]} {
	    $wThumb configure -text $img
	}
    }

    proc GetAndUpdateThumbnail { x y w tbl wThumb} {

	# set tablelist::x [expr {$x + [winfo x $w]}]
	# set tablelist::y [expr {$y + [winfo y $w]}]
	# set cell [$tbl nearestcell $tablelist::x $tablelist::y]
	# foreach {row col} [ split $cell ,] break    

	set setsel [$tbl curselection]

	set model_unique_name "VELaSSCo_Models:*:*"
	if { $setsel == {} } {
	    tk_dialogRAM $w.terr error [_ "Sorry, no model selected."] error 0 [_ "OK"]
	} else {
	    if { [ llength $setsel] == 1} {
		set row_values [ $tbl get $setsel]
		set model_name [ lindex $row_values 0]
		set model_path [ lindex $row_values 1]
		set model_loc [ lindex $row_values 2]
		# location can be of the form HBase:TableName or EDM
		if { [ string compare -nocase "EDM" $model_loc] == 0 } {
		    # it matches "EDM"
		    set model_unique_name $model_name
		} else {
		    # if location == HBASE:TableName then location has the table name
		    # otherwise don't use it, as we don't know
		    set location_lst [ split $model_loc :]
		    set location_lst_size [ llength $location_lst]
		    if { ( $location_lst_size >= 1) && [ string equal -nocase [ lindex $location_lst 0] "hbase"]} {
			if { $location_lst_size == 2} {
			    set model_unique_name "[ lindex $location_lst 1]:$model_path:$model_name"
			} else {
			    set model_unique_name "$model_path:$model_name"
			}
		    } else {
			# for non-hbase storage module, i.e. EDM, use only the name as unique_model_name
			set model_unique_name "$model_name"
		    }
		}
	    }
	}
	# at the moment no GetThumbnailOfModel query
	# Debug "VQuery GetThumbnailOfModel( $model_unique_name)"
	# after 500 [ list VELaSSCo::DoUpdateThumbImage $wThumb [_ "Showing thumbnail for model '%s'" $model_unique_name]]
	set ::VELaSSCo(SelectedModel) $model_unique_name
    }

    proc FillModelTable { tbl} {
	variable _sessionID

	$tbl delete 0 end

	set group $::VELaSSCo(group_qualifier)
	set name $::VELaSSCo(name_pattern)
	set ret [ gid_velassco GetListOfModels $_sessionID $group $name]
	# Debug "gid_velassco GetListOfModels $_sessionID $group $name = \n$ret"

	# filling list
	set lst_names {}
	set lst_fp {}
	set lst_model_ids {}
	set lst_locations {}
	
	# we skip the first line: "NumberOfModels: N"
	set lst_models [ lrange [ split $ret "\n"] 1 end]
	foreach entry $lst_models {
	    set keyword [ lindex $entry 0]
	    set value [ lrange $entry 1 end]
	    switch $keyword {
		"Name:" {
		    lappend lst_names $value
		}
		"FullPath:" {
		    lappend lst_fp $value
		}
		"ModelID:" {
		    lappend lst_model_ids $value
		}
		"Location:" {
		    lappend lst_locations $value
		}
	    }
	}
	foreach name $lst_names fp $lst_fp lc $lst_locations modelID $lst_model_ids {
	    $tbl insert end [ list $name $fp $lc $modelID]
	    # after 100
	}
    }

    proc SelectModel { { w .gid.wVelasscoSelectModel}} {
	if { ![ CheckLogin]} { return}
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo model selection..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Open"] [_ "Cancel"]] \
	    [ list [ list VELaSSCo::DoSelectModel $w] [ list destroy $w]] \
	    [ list <Return> <Escape>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo model selection"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	package require tablelist_tile
	ttk::frame $w.top.selection

	ttk::frame $w.top.selection.filter
	ttk::label $w.top.selection.filter.lgroup -text [_ "Group qualifier:"]
	ttk::entry $w.top.selection.filter.egroup -width 20 -textvariable ::VELaSSCo(group_qualifier)
	ttk::label $w.top.selection.filter.lpattern -text [_ "Name pattern:"]
	ttk::entry $w.top.selection.filter.epattern -width 20 -textvariable ::VELaSSCo(name_pattern)
	ttk::button $w.top.selection.filter.brefresh -text [_ "Refresh table"]
	grid $w.top.selection.filter.lgroup $w.top.selection.filter.egroup \
	     $w.top.selection.filter.lpattern $w.top.selection.filter.epattern \
	    $w.top.selection.filter.brefresh \
	    -padx 4 -pady 4 -sticky e
	grid configure  $w.top.selection.filter.egroup -sticky w
	grid configure  $w.top.selection.filter.epattern -sticky w
	grid configure  $w.top.selection.filter.brefresh -sticky w

	set tbl_sw [ ScrolledWindow $w.top.selection.sw -auto both -relief solid -borderwidth 0]
	set tbl $w.top.selection.sw.tbl
	tablelist::tablelist $tbl \
	    -stretch all -background white \
	    -columns [ list 20 [_ "Name"] 0 [_ "Full path"] 0 [_ "Location"] 32 [_ "Model ID"]]
	    # -columns [ list 30 [_ "Name"] 0 [_ "Full path"] 20 [_ "Validation status"] 20 [_ "Access"]]

	$w.top.selection.sw setwidget $tbl

	ttk::label $w.top.selection.thumb -text [_ "Thumbnail image of selected model"]
	bind $tbl <<TablelistSelect>> [ list VELaSSCo::GetAndUpdateThumbnail %x %y %W $tbl $w.top.selection.thumb]

	bind [$tbl bodytag] <Double-Button-1> [ list VELaSSCo::DoSelectModel $w]

	grid $w.top.selection.filter -sticky ew
	# grid $w.top.selection.tbl -padx 4 -pady 4 -sticky news
	grid $w.top.selection.sw -padx 4 -pady 4 -sticky news
	grid $w.top.selection.thumb  -padx 4 -pady 4 -sticky ew

	grid columnconfigure $w.top.selection 0 -weight 1
	grid rowconfigure $w.top.selection 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.selection -sticky ewns
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1
	wm geometry $w 990x325

	set ::VELaSSCo(group_qualifier) "*"
	set ::VELaSSCo(name_pattern) "*"

	$w.top.selection.filter.brefresh configure -command [ list ::VELaSSCo::FillModelTable $tbl] 
	FillModelTable $tbl
    }

    proc AlreadyInGiD { selected_analysis selected_step res_name} {
	variable _cache
	set in_gid 0
	if { [ info exists _cache(InGiD,Result,$selected_analysis,$selected_step,$res_name)] && \
		 $_cache(InGiD,Result,$selected_analysis,$selected_step,$res_name)} {
	    set in_gid 1
	}
	return $in_gid
    }

    proc CreateResultInGiD { selected_analysis selected_step res_name lst_ids lst_values} {
	variable _cache
	# res_header = [ list Result $res_name $selected_analysis $selected_step $res_type $location]
	set res_header $_cache($selected_analysis,$selected_step,$res_name,Header)
	set res_comp $_cache($selected_analysis,$selected_step,$res_name,ComponentNames)
	set res_type [ lindex $res_header 4]
	set num_comp $_cache($selected_analysis,$selected_step,$res_name,NumberOfComponents)
	
	set lst_gid_res {}
	for { set idx_id 0} { $idx_id < [ llength $lst_ids]} { incr idx_id} {
	    set node_id [ lindex $lst_ids $idx_id]
	    set value [ lindex $lst_values $idx_id]
	    lappend lst_gid_res [ list $node_id {*}$value]
	}
	# set lst_gid_res [ list $lst_ids $lst_values]
	if { [ llength $res_comp] == 0} {
	    GiD_Result create $res_header {*}$lst_gid_res
	} else {
	    set components [ list ComponentNames {*}$res_comp]
	    GiD_Result create $res_header $components {*}$lst_gid_res
	}
	set _cache(InGiD,Result,$selected_analysis,$selected_step,$res_name) 1
    } 

    proc CreateResultTree { selected_analysis lst_steps lst_results} {
	variable _selected
	variable _cache
	# assuming the results are the same for all time-steps:
	set lst_results [ split $lst_results \n]
	set num_res 0
	set read_num_res 0
	set _cache(ListOfResults) {}
	foreach pair $lst_results {
	    set key [ lindex $pair 0]
	    if { $key == ""} {
		continue
	    }
	    switch $key {
		"NumberOfResults:" {
		    set read_num_res [ lindex $pair 1]
		}
		"Name:" {
		    # new result
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		    lappend _cache(ListOfResults) [ lrange $pair 1 end]
		}
		"NumberOfComponents:" {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
		"ComponentNames:" {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
		"ResultNumber:" {
		    # last result info field
		    set resInfo($num_res,$key) [ lindex $pair 1]
		    incr num_res
		}
		default {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
	    }
	}
	
	# lets create dummy results in order to have the result menu's...
	# GiD_Result needs a mesh first ...
	# GiD_MeshPost create "dummyMesh" Point 1 {1} {0.0 0.0 0.0} {1} {1}
	# GiD_MeshPost create "dummyMesh" Sphere 1 {1} {0.0 0.0 0.0} {1} {1} {0.0}

	if { [ llength [ array names _cache InGiD,Mesh,*]] == 0} {
	    GiD_MeshPost create "dummyMesh" Sphere 1 -zero_based_array {0} {0.0 0.0 0.0} {0} {0} {0.0}
	    set _cache(InGiD,Mesh,dummyMesh) 1
	}

	set _cache(VerticesID) {1}
	foreach selected_step $lst_steps {
	    for { set ires 0} { $ires < $read_num_res} { incr ires} {
		set res_name $resInfo($ires,Name:)
		# analysis already selected
		# step already selected, eventually loop over all steps to create the results
		set res_type $resInfo($ires,ResultType:)
		# id_node result_value(s)
		set dummy_result {1 0}
		switch $res_type {
		    Scalar { set dummy_result {1 0}}
		    Vector { set dummy_result {1 0 0 0}}
		    Matrix { set dummy_result {1 0 0 0 0 0 0}}
		}
		set location $resInfo($ires,Location:)
		set res_header [ list Result $res_name $selected_analysis $selected_step $res_type $location]
		if { [ string equal -nocase $location "OnGaussPoints"]} {
		    lappend res_header $resInfo($ires,GaussPointName:)
		}
		if { [ llength $resInfo($ires,ComponentNames:)] == 0} {
		    GiD_Result create $res_header $dummy_result
		} else {
		    set components [ list ComponentNames {*}$resInfo($ires,ComponentNames:)]
		    GiD_Result create $res_header $components $dummy_result
		}
		# W "created  $res_header"
		set _cache($selected_analysis,$selected_step,$res_name,Header) $res_header
		set _cache($selected_analysis,$selected_step,$res_name,Type) $res_type
		set _cache($selected_analysis,$selected_step,$res_name,ComponentNames) $resInfo($ires,ComponentNames:)
		set _cache($selected_analysis,$selected_step,$res_name,NumberOfComponents) $resInfo($ires,NumberOfComponents:)
	    }
	}

	# select the last one:
	GiD_Process Mescape results analysissel [ toGiDEncoding $selected_analysis] $selected_step
	set _selected(analysis) $selected_analysis
	set _selected(step) $selected_step
    }

    proc GetModelInfo { show_info} {
	variable _sessionID
	variable _modelID
	variable _cache
	variable _selected
	if { ![ CheckLoginAndModelID]} { return}
	set doing ""
	set err [ catch {
	    # testing GetListOfAnalyses
	    set doing "gid_velassco GetListOfAnalyses $_sessionID $_modelID"
	    set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
	    set selected_analysis [ lindex $lst_analyses 0]
	    if { [ llength $lst_analyses] > 1} {
		# if more than one, select the last one
		set selected_analysis [ lindex $lst_analyses end-1]
	    }
	    set _selected(analysis) $selected_analysis
	    set _cache(ListOfAnalyses) $lst_analyses

	    if { $show_info} {
		W "gid_velassco GetListOfAnalyses $_sessionID $_modelID = "
		W $lst_analyses
		W "  selected = $selected_analysis"
		W \n
	    }
	    # testing GetListOfSteps
	    set doing "gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis"
	    set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis]
	    set selected_step [ lindex $lst_steps 0]
	    if { [ llength $lst_steps] > 1} {
		# if more than one, select the last one
		set selected_step [ lindex $lst_steps end-1]
	    }
	    set _selected(step) $selected_step
	    set _cache(ListOfSteps) $lst_steps
	    set num_steps [ llength $lst_steps]
	    if { $show_info} {
		W "gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis = "
		W "Number of Steps = $num_steps"
		if { $num_steps < 20} {
		    W $lst_steps
		} else {
		    # show only some of them
		    W "[ lrange $lst_steps 0 7] . . . [ lrange $lst_steps [ expr $num_steps - 8] end]"
		}
		W "  selected = $selected_step"
		W \n
	    }
	    # testing GetListOfResults
	    set doing "gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step"
	    set lst_results [ gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step]
	    if { $show_info} {
		W "gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step = "
		W $lst_results
		W \n
	    }

	    # testing GetListOfMeshes
	    set gm_analysis $selected_analysis
	    set gm_step $selected_step
	    if { "$selected_analysis" != "DEM"} {
		# get information about static meshes
		set gm_analysis ""
		set gm_step 0.0
	    }
	    set doing "gid_velassco GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step"
	    set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step]
	    if { $show_info} {
		W "gid_velassco GetListOfMeshes $_sessionID $_modelID \"$gm_analysis\" $gm_step = "
		W $lst_meshes
		W \n
	    }

	    set _cache(GetListOfMeshes,Query) "GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step"
	    set _cache(GetListOfMeshes,Results) $lst_meshes

	} err_txt]
	
	if { $err} {
	    W "ERROR doing $doing"
	    W $err_txt
	} else {
	    # delete any previous mesh and results info
	    # GiD_Process Mescape Files New Yes escape
	    # CreateResultTree $selected_analysis $selected_step $lst_results
	    CreateResultTree $selected_analysis $lst_steps $lst_results
	}
    }

    proc DoTunnelLinux {} {

	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set err [ catch {
	    set fo [ file tempfile tmp_file velassco]
	    puts $fo "#!/usr/bin/expect"
	    puts $fo "# Using ssh from expect"
	    puts $fo "log_user 0"
	    puts $fo "spawn ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	    puts $fo "expect \"sword: \""
	    puts $fo "send \"$::VELaSSCo(Tunnel,Password)\\r\""
	    puts $fo "interact"
	    close $fo
	    file attributes $tmp_file -permissions u+x,go-rwx
	    set _tunnel_used [ exec $tmp_file &]
	    # Debug $_tunnel_used
	    after 1000
	    file delete -force $tmp_file
	} err_txt]
	if { $err} {
	    Print "Could not use expect: '$err_txt', using interactive xterm isntead."
	    set _tunnel_used [ exec xterm -e ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	    after 5000
	}
    }

    proc DoTunnelWindows {} {
	variable _dir
	variable _tunnel_used
	
	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set bin_putty [ file join $_dir bin putty.exe]
	set _tunnel_used [ exec "$bin_putty" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# set putty_cmd "\"$bin_putty\" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	# proc ::___dummy__cb {} {
	# }
	# set _tunnel_used [ runExe run $putty_cmd -blocking false -callback ::___dummy__cb -timeout 0]
	after 5000
    }

    proc DoLogin { w} {
	variable _sessionID
	variable _tunnel_used

	if { $_sessionID != ""} {
	    WarnWin [_ "Already logged in!"]
	    return
	}

	if { $::VELaSSCo(Tunnel,enabled)} {

	    if { $::tcl_platform(platform) == "windows"} {
		DoTunnelWindows
	    } else {
		DoTunnelLinux
	    }

	    set gv_id [ gid_velassco UserLogin localhost $::VELaSSCo(Tunnel,LocalPort) $::VELaSSCo(Username) $::VELaSSCo(Password)]

	    # when closing the window do a logout ...
	    bind $w <Destroy> "+::VELaSSCo::DoLogout"

	} else {
	    set _tunnel_used 0
	    set gv_id [ gid_velassco UserLogin $::VELaSSCo(Host) $::VELaSSCo(Port) $::VELaSSCo(Username) $::VELaSSCo(Password)]
	}
	set _sessionID $gv_id

	ShowLoginStatus
    }

    proc DoLogout {} {
	variable _sessionID
	variable _tunnel_used
	if { $_sessionID != ""} {
	    set ret ""
	    catch { 
		set ret [ gid_velassco UserLogout $_sessionID]
	    }
	    if { $_tunnel_used != 0} {
		# kill tunnel
		if { $::tcl_platform(platform) == "windows"} {
		    # runExe stop $_tunnel_used
		    # needed to kill the tunnel
		    package require gid_cross_platform
		    # can not do the package require when needed because sometimes gid is being destroyed
		    gid_cross_platform::end_process $_tunnel_used
		} else {
		    exec kill $_tunnel_used
		}
	    }
	    Verbose 1 "gid_velassco UserLogout $_sessionID = $ret"
	}
	set _sessionID ""
	set _tunnel_used 0
	ShowLoginStatus
    }

    proc DoStatusDB {} {
	variable _sessionID
	if { ![ CheckLogin]} { return}
	InfoWin [_ "gid_velassco GetStatusDB $_sessionID = \n%s" \
		     [ gid_velassco GetStatusDB $_sessionID]]
    }

    proc DoStopVELaSSCo {} {
	variable _sessionID
	if { ![ CheckLogin]} { return}
	InfoWin [_ "gid_velassco StopVELaSSCo $_sessionID = \n%s" \
		     [ gid_velassco StopVELaSSCo $_sessionID]]
	DoLogout
    }

    proc EnableDisableTunnelWidgets { lst_w_tunnel} {
	foreach w $lst_w_tunnel {
	    set state disabled
	    if { $::VELaSSCo(Tunnel,enabled)} {
		set state normal

		# set new default parameters
		set ::VELaSSCo(Host) pez001
		set ::VELaSSCo(Port) 26267
	    }
	    $w configure -state $state
	}
    }

    proc ShowLoginStatus {} {
	variable _sessionID
	variable _status_wid
	variable _login_icon_wid
	variable _icon_do_connection
	variable _icon_connected
	
	if { ( $_status_wid != "") && [ winfo exists $_status_wid]} {
	    set color green
	    set txt "Logged in"
	    if { $_sessionID == ""} {
		set color red
		set txt "Not logged in"
	    }
	    $_status_wid configure -background $color
	    $_status_wid.txt configure -text $txt -background $color -foreground white
	}

	if { ( $_login_icon_wid != "") && [ winfo exists $_login_icon_wid]} {
	    set img $_icon_connected
	    if { $_sessionID == ""} {
		set img $_icon_do_connection
	    }
	    $_login_icon_wid configure -image [ VELaSSCo::GetImage $img]
	}
    }

    proc CheckLogin {} {
	variable _sessionID

	set ret 1
	if { $_sessionID == ""} {
	    WarnWin [_ "Please, log in first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }
    proc CheckLoginAndModelID {} {
	variable _modelID

	if { ![ CheckLogin]} { return 0}
	set ret 1
	if { $_modelID == ""} {
	    WarnWin [_ "Please, open a model first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }

    proc OGLdrawLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    foreach name [ array names _ogl list,*] {
		GiD_OpenGL draw -calllist $_ogl($name)
	    }
	}
    }
    proc OGLregisterDraw {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    set _ogl(register_id) [ GiD_OpenGL register ::VELaSSCo::OGLdrawLists]
	}
    }
    proc OGLunregisterAndDeleteLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    if { [ info exists _ogl(register_id)] && ( $_ogl(register_id) != -1)} {
		GiD_OpenGL unregister $_ogl(register_id)
		set _ogl(register_id) -1
	    }
	    # now deleting lists
	    foreach name [ array names _ogl list,] {
		GiD_OpenGL draw -deletelists $_ogl($name) 1
	    }
	}
    }

    proc OGLcreateBBoxList { bbox} {
	variable _ogl
	
	# create list
	set _ogl(list,bbox) [ GiD_OpenGL draw -genlists 1]
	foreach var [ list min_x min_y min_z max_x max_y max_z] val $bbox {
	    set $var $val
	}
	
	set vertices(0) [ list $min_x $min_y $min_z]
	set vertices(1) [ list $max_x $min_y $min_z]
	set vertices(2) [ list $max_x $max_y $min_z]
	set vertices(3) [ list $min_x $max_y $min_z]
	set vertices(4) [ list $min_x $min_y $max_z]
	set vertices(5) [ list $max_x $min_y $max_z]
	set vertices(6) [ list $max_x $max_y $max_z]
	set vertices(7) [ list $min_x $max_y $max_z]
	set lst_lines [ list \
			    [ list 0 1] [ list 1 2] [ list 2 3] [ list 3 0] \
			    [ list 0 4] [ list 1 5] [ list 2 6] [ list 3 7] \
			    [ list 4 5] [ list 5 6] [ list 6 7] [ list 7 4]]

	GiD_OpenGL draw -newlist $_ogl(list,bbox) compile
	# somewhat dark magenta
	GiD_OpenGL draw -color "0.69841 0.190464 0.69841"
	GiD_OpenGL draw -begin lines
	foreach ln $lst_lines {
	    foreach nd $ln {
		GiD_OpenGL draw -vertex $vertices($nd)
	    }
	}
	# end lines
	GiD_OpenGL draw -end
	# end of  list
	GiD_OpenGL draw -endlist

	#register OpenGL proc if not already registered
	OGLregisterDraw
    }

    proc GIDcreateBoundaryMeshFromObjArray { mesh_obj mesh_name element_type num_nodes_per_element} {
	variable _cache

	# returns a list of:
	# list_of_vertices = list of id's + list of coords
	# list_of_triangle_nodes
	# example: { { { 1 2 3} { 0.0 0.0 0.0 1.0 0.0 0.0 0.0 1.0 0.0}} { 1 2 3}}
	set vertices [ lindex $mesh_obj 0]
	set triangles [ lindex $mesh_obj 1]
	set lst_node_ids [ lindex $vertices 0]
	set lst_coords [ lindex $vertices 1]
	set _cache(VerticesID) $lst_node_ids
	# still there is one convertion UpdateStringFromLongLongArray

	# delete dummy mesh used to create result menus...
	set VersionRequired "12.1.10"
	set comp -1
	catch { 
	    set comp [ GiDVersionCmp $VersionRequired]
	}
	if { $comp == -1} {
	    # GiD_MeshPost has a bug in versions 12.1.9 and lower, so vertices ID needs a conversion:
	    set zeroed_id_list [ lmap id $lst_node_ids { expr $id - 1}]
	    set zeroed_triangle_list [ lmap id $triangles { expr $id - 1}]
	    # but using this conversion transforms the objarray into a TclString ... double the memory required
	    GiD_MeshPost create $mesh_name $element_type $num_nodes_per_element \
		-zero_based_array $zeroed_id_list $lst_coords {} $zeroed_triangle_list
	    set _cache(InGiD,Mesh,$mesh_name) 1
	} else {
	    GiD_MeshPost create $mesh_name $element_type $num_nodes_per_element \
		$lst_node_ids $lst_coords {} $triangles
	    set _cache(InGiD,Mesh,$mesh_name) 1
	}
	GiD_Process Mescape Utilities Delete VolumeSets "dummyMesh" Yes escape escape
	set _cache(InGiD,Mesh,dummyMesh) 0
    }

    proc OGLcreateBoundaryMeshListFromObjArray { mesh_obj} {
	variable _ogl
	variable _modelInfo
	
	# create list
	set _ogl(list,boundaryMesh) [ GiD_OpenGL draw -genlists 1]
	set vertices [ lindex $mesh_obj 0]
	set triangles [ lindex $mesh_obj 1]
	set lst_node_ids [ lindex $vertices 0]
	set lst_coords [ lindex $vertices 1]

	# there are convertions
	#  UpdateStringFromLongLongArray and
	#  UpdateStringFromDoubleArray
	# which can be avoided using 
	#  objarray get $lst_node_ids $idx
	#  objarray get $lst_coords $idx
	#  objarray get $triangles $idx
	# or implementing a new function
	#   objarray map tri_node $lst_triangles {GiD_OpenGL draw -vertex $tri_node}
	#   line lmap

	set min_x    1e+38
	set max_x   -1e+38
	set min_y    1e+38
	set max_y   -1e+38
	set min_z    1e+38
	set max_z   -1e+38
	foreach id $lst_node_ids [ list x y z] $lst_coords {
	    set vertex_map($id) [ list $x $y $z]
	    if { $x < $min_x} { set min_x $x}
	    if { $x > $max_x} { set max_x $x}
	    if { $y < $min_y} { set min_y $y}
	    if { $y > $max_y} { set max_y $y}
	    if { $z < $min_z} { set min_z $z}
	    if { $z > $max_z} { set max_z $z}
	}
	# define the zoom_frame box
	if { ![ info exists _modelInfo] || ![ info exists _modelInfo(bbox)]} {
	    set _modelInfo(bbox) [ list $min_x $min_y $min_z $max_x $max_y $max_z]
	}

	GiD_OpenGL draw -newlist $_ogl(list,boundaryMesh) compile
	# somewhat dark green
	GiD_OpenGL draw -color "0.0 0.0 0.0"
	GiD_OpenGL draw -begin lines
	foreach [ list n1 n2 n3] $triangles {
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	}

	GiD_OpenGL draw -end

	GiD_OpenGL doscrzoffset true
	GiD_OpenGL draw -color "0.8 0.8 0.6"
	GiD_OpenGL draw -begin triangles
	foreach [ list n1 n2 n3] $triangles {
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	}
	GiD_OpenGL draw -end
	GiD_OpenGL doscrzoffset false

	# end of  list
	GiD_OpenGL draw -endlist

	#register OpenGL proc if not already registered
	OGLregisterDraw
    }


    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    proc TclCalcModelBoundaries { useof} {
	variable _modelInfo
	if { [ info exists _modelInfo] && [ info exists _modelInfo(bbox)]} {
	    return $_modelInfo(bbox)
	} else {
	    # return [ list 1e+38 1e+38 1e+38 -1e+38 -1e+38 -1e+38]
	    # return an empty list, otherwise it will use it and not the internal one!
	    return {}
	}
    }

    proc DoGetBoundingBox {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLoginAndModelID]} { return}
	WaitState
	Verbose 1 "calling gid_velassco GetBoundingBox $_sessionID $_modelID"
	set modelBBox [ gid_velassco GetBoundingBox $_sessionID $_modelID]
	# set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	Verbose 1 "selected model = BBox = $modelBBox"
	
	# register a drawopengl procedure to draw the bounding box
	if { [ llength $modelBBox] == 6} {
	    OGLcreateBBoxList $modelBBox
	}
	set _modelInfo(bbox) $modelBBox
	GiD_Process 'Zoom Frame
	EndWaitState
    }

    proc GetListOfTetrahedraMeshes {} {
	variable _sessionID
	variable _modelID
	variable _cache

	if { ![ info exists _cache(GetListOfMeshes,Results)]} {
	    Verbose 1 "getting Mesh information list"
	    # get the list of meshes
	    # testing GetListOfAnalyses
	    set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
	    set analysis [ lindex $lst_analyses 0]
	    # testing GetListOfSteps
	    set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $analysis]
	    set step [ lindex $lst_steps 0]
	    
	    # testing GetListOfMeshes
	    if { "$analysis" != "DEM"} {
		# get information about static meshes
		set analysis ""
		set step 0.0
	    }
	    set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $analysis $step]

	    set _cache(GetListOfMeshes,Query) "GetListOfMeshes $_sessionID $_modelID $analysis $step"
	    set _cache(GetListOfMeshes,Results) $lst_meshes
	}

	# the information of the meshes is like this:
	#   NumberOfMeshes: 11
	#   Name: volumes
	#   ElementType: Tetrahedra
	#   NumberOfVerticesPerElement: 0
	#   NumberOfVertices: 3391
	#   NumberOfElements: 656399645689249792
	#   Units: 
	#   Color: 
	#   MeshNumber: 1
	#   CoordsName: 
	#   Name: cil
	#   ElementType: Tetrahedra
	#   NumberOfVerticesPerElement: 0
	#   NumberOfVertices: 0
	#   NumberOfElements: 960
	#   Units: 
	#   Color: 
	#   MeshNumber: 2
	#   CoordsName: 
	#   Name: wake
	#   ...
	set lst_tetra_meshes {}
	set lst_MeshInfo [ split $_cache(GetListOfMeshes,Results) "\n"]
	foreach mesh_info $lst_MeshInfo {
	    set key [ lindex $mesh_info 0]
	    if { $key == "Name:"} {
		set last_name [ lrange $mesh_info 1 end]
	    } elseif { $key == "ElementType:"} {
		set elType [ lindex $mesh_info 1]
		if { $elType == "Tetrahedra"} {
		    lappend lst_tetra_meshes $last_name
		}
	    }
	}
	return $lst_tetra_meshes
    }

    proc DoGetBoundaryOfAMesh {} {
	variable _sessionID
	variable _modelID
	variable _cache

	if { ![ CheckLoginAndModelID]} { return}

	WaitState
	foreach selected_mesh_name [ GetListOfTetrahedraMeshes] {
	    Verbose 1 "calling gid_velassco GetBoundaryOfAMesh $_sessionID $_modelID $selected_mesh_name"
	    set mesh_data [ gid_velassco GetBoundaryOfAMesh $_sessionID $_modelID $selected_mesh_name]
	    # set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	    set vertex_info [ lindex $mesh_data 0]
	    set vertex_ids [ lindex $vertex_info 0]
	    set triangle_nodes [ lindex $mesh_data 1]
	    set num_vertices [ objarray length $vertex_ids]
	    set num_triangles [ expr [ objarray length $triangle_nodes] / 3]
	    Verbose 1 " got boundary of '$selected_mesh_name' with $num_vertices vertices and $num_triangles faces"
	    
	    # it should have create a GiD_PostMesh but in the meantine ...
	    # register a drawopengl procedure to draw the boundary mesh
	    if { [ llength $mesh_data] == 2} {
		# GIDcreateBoundaryMesh needs a correction in GiD 12.1.10
		# set VersionRequired "12.1.10"
		# set comp -1
		# catch { 
		#     # set comp [ GiDVersionCmp $VersionRequired]
		# }
		# update
		# if { $comp == -1 } {
		#     # use tcl drawopengl fallback.
		#     OGLcreateBoundaryMeshListFromObjArray $mesh_data
		# } else {
		    GIDcreateBoundaryMeshFromObjArray $mesh_data "$selected_mesh_name boundary" Triangle 3
		# }
	    }
	    GiD_Process 'Zoom Frame
	}
	EndWaitState
    }

    proc Login { { w_bar_icon ""} { w .gid.wVelasscoLogin}} {
	variable _status_wid
	variable _login_icon_wid
	if { $w_bar_icon != ""} {
	    set _login_icon_wid $w_bar_icon
	}

	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo access window..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Login"] [_ "Status"] [_ "Logout"]] \
	    [ list [ list VELaSSCo::DoLogin $w] \
		  VELaSSCo::DoStatusDB VELaSSCo::DoLogout] \
	    [ list <Return> ""]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo login"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.login
	foreach id [ list 1 2 3 4] \
	    var_name [ list "Username" "Password" "Host" "Port"] \
	    txt [ list [_ "Username"] [_ "Password"] [_ "VELaSSCo host"] [_ "Port"]] \
	    width [ list 16 16 20 8] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo($var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	ttk::checkbutton $w.top.login.cbTunel -text [_ "Tunnel through"]: \
	    -variable ::VELaSSCo(Tunnel,enabled)
	grid $w.top.login.cbTunel -padx 4 -pady 4 -sticky w -columnspan 2
	set lst_w_tunnel {}
	foreach id [ list 5 6 7 8] \
	    var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	    txt [ list [_ "Tunnel host"] [_ "local port"] [_ "Username"] [_ "Password"] ] \
	    width [ list 20 8 16 16] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo(Tunnel,$var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		lappend lst_w_tunnel $w.top.login.l$id $w.top.login.e$id
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	$w.top.login.cbTunel configure -command [ list VELaSSCo::EnableDisableTunnelWidgets $lst_w_tunnel]

	frame $w.top.status -height 32 -bd 0
	ttk::label $w.top.status.txt -justify center
	grid $w.top.status.txt -padx 4 -pady 4
	
	set _status_wid $w.top.status

	grid columnconfigure $w.top.login 0 -weight 1
	grid columnconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 2 -weight 1

	grid columnconfigure $w.top.status 0 -weight 1
	grid rowconfigure $w.top.status 0 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.login -sticky ew
	grid $w.top.status -sticky ew
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# default values
	# foreach var_name [ list "Username" "Password" "Host" "Port"] \
	#     value [ list miguel 1234velassco pez001 9990] {
	# 	set ::VELaSSCo($var_name) $value
	#     }
	# set ::VELaSSCo(Tunnel,enabled) 1
	# foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	#     value [ list acuario.cimne.upc.edu 9990 miguel 1234] {
	# 	set ::VELaSSCo(Tunnel,$var_name) $value
	#     }
	# set ::VELaSSCo(SelectedModel) ""
	foreach var_name [ list "Username" "Password" "Host" "Port"] \
	    value [ list andreas 1234 localhost 26267] {
		set ::VELaSSCo($var_name) $value
	    }
	set ::VELaSSCo(Tunnel,enabled) 0
	foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	    value [ list acuario.cimne.upc.edu 9990 miguel ****] {
		set ::VELaSSCo(Tunnel,$var_name) $value
	    }
	set ::VELaSSCo(SelectedModel) ""

	# enable/disable tunel widgets ...
	VELaSSCo::EnableDisableTunnelWidgets $lst_w_tunnel
	ShowLoginStatus
    }

    proc GetResultForVertices { result_name} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _busy
	if { ![ CheckLoginAndModelID]} { return}	

	# get result values from VELaSSCo and
	# updates them into GiD:
	#  GetResultFromVerticesID sessionID modelID resultID analysisID lst_vertexIDS timeStep

	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set res_type $_cache($sel_analysis,$sel_step,$result_name,Type)
	set res_num_comp $_cache($sel_analysis,$sel_step,$result_name,NumberOfComponents)

	Verbose 1 "GetResultFromVerticesID $_sessionID $_modelID \"$result_name\" \"$_selected(analysis)\" 1 $_selected(step)"
	if { [ info exists _cache(VerticesID)] && ( [ llength $_cache(VerticesID)] != 0)} {
	    set res_values [ gid_velassco GetResultFromVerticesID $_sessionID $_modelID \
				 "$result_name" $res_type $res_num_comp \
				 "$_selected(analysis)" $_cache(VerticesID) $_selected(step)]
	    # returns 2 lists: lst_ids and lst_results
	    # W "lst_ids size = [ llength [ lindex $res_values 0]]"
	    # W "lst_res size = [ llength [ lindex $res_values 1]]"
	}
	return $res_values
    }

    proc GetResultForNodeEvolution { result_name lst_vertex_ids} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _busy
	if { ![ CheckLoginAndModelID]} { return}	

	# get result values from VELaSSCo and
	# updates them into GiD:
	#  GetResultFromVerticesID sessionID modelID resultID analysisID lst_vertexIDS timeStep

	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set res_type $_cache($sel_analysis,$sel_step,$result_name,Type)
	set res_num_comp $_cache($sel_analysis,$sel_step,$result_name,NumberOfComponents)

	set res_values {}
	if { [ info exists _cache(ListOfSteps)] && ( [ llength $_cache(ListOfSteps)] >= 1) } {
	    # the vquery does not accept a list of steps as a parameter ...
	    foreach step $_cache(ListOfSteps) {
		Verbose 1 "GetResultFromVerticesID $_sessionID $_modelID \"$result_name\" \"$_selected(analysis)\" $res_type $res_num_comp $lst_vertex_ids $step"
		set result_ids_values [ gid_velassco GetResultFromVerticesID $_sessionID $_modelID \
					    "$result_name" $res_type $res_num_comp \
					    "$_selected(analysis)" $lst_vertex_ids $step]
		set graf_ids [ lindex $result_ids_values 0]
		set graf_values [ lindex $result_ids_values 1]
		# although it's only one value it's a nested list: { {0.0 0.0 0.0}}
		lappend res_values [ lindex $graf_values 0]
	    }
	} else {
	    WarnWin [_ "Please, get the model information first."]
	}
	# returns 2 lists: lst_ids and lst_results
	# W "lst_ids size = [ llength [ lindex $res_values 0]]"
	# W "lst_res size = [ llength [ lindex $res_values 1]]"
	return $res_values
    }

    proc doNodeEvolution { action res_name comp_name name1 name2 op} {
	variable _cache
	variable _selected
	trace remove variable ::GidPriv(selection) write [ list ::VELaSSCo::catchDoNodeEvolution $action $res_name $comp_name]
	# set res_name [ lindex $_cache(ListOfResults) 0]
	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set idx_comp 0
	if { $comp_name != ""} {
	    set idx_comp [ lsearch $_cache($sel_analysis,$sel_step,$res_name,ComponentNames) $comp_name]
	    if { $idx_comp == -1} {
		set idx_comp 0
	    }
	}

	# $::GidPriv(selection) is of the form "Nodes 1" (with the quotes)
	set node_selection [ regsub -all \" $::GidPriv(selection) { }]
	set node_id [ lindex $node_selection 1]
	set graf_values [ GetResultForNodeEvolution $res_name $node_id]
	# W "graf_values = $graf_values"

	# create a graph file:
	set tmp_filename [ file join [ GiD_Info project TmpDirectory] "Evolution_${res_name}_node_${node_id}.grf"]
	set fo [ open $tmp_filename w]
	puts $fo "# Graph: \"Evolution of $res_name on node $node_id\""
	puts $fo "# X: \"Step value\" Y: \"$res_name\""
	foreach step $_cache(ListOfSteps) value $graf_values {
	    puts $fo "$step [ lindex $value $idx_comp]"
	}
	close $fo
	# now read the graph in GiD
	GiD_Process Mescape Results Graphs ReadGraph "$tmp_filename" escape
	# W "file $tmp_filename created"
	# processing GiD_Process takes some time and severl Tcl event loops ...
	after 3000 [ list file delete -force $tmp_filename]
    }

    proc catchDoNodeEvolution { action res_name comp_name name1 name2 op} {
	set err [
		 catch {
		     doNodeEvolution $action $res_name $comp_name $name1 $name2 $op
		 } err_txt
		 ]
	if { $err} {
	    W "doNodeEvolution $action $res_name $comp_name $name1 $name2 $op = \n\t$err_txt"
	}
    }

    proc startNodeEvolution { result_name { component_name ""}} {
	variable _cache
	if { ![ info exists _cache(ListOfSteps)] || ( [ llength $_cache(ListOfSteps)] < 1) } {
	    WarnWin [_ "Please, get the model information first."]
	    return
	}
	GiD_Process 'SelectEntities JustOne Nodes
	trace add variable ::GidPriv(selection) write [ list ::VELaSSCo::catchDoNodeEvolution NodeEvolution \
							    [ fromGiDEncoding $result_name] [ fromGiDEncoding $component_name]]
    }


    proc AddGraphsResultsToMenu { w basecommand que_res } {
	global VelasscoViewMenuPriv
	
	# registramos los menus
	# if { [ esMac]} {
	#         registerPostGraphsResultsMenu $w $basecommand $que_res
	#         $w configure -postcommand ""
	# }
	
	set all_anal [lsort -dictionary [ GiD_Info postprocess get all_analysis]]
	# set all_anal [ GiD_Info postprocess get all_analysis]
	if { $all_anal == ""} {
	    set msg [_ "Please load a MODEL first"]
	} else {
	    set msg [_ "Please select a STEP first"]
	}
	set cur_analysis [ GiD_Info postprocess get cur_analysis]
	if { $cur_analysis == "" } {
	    $w delete 0 end
	    $w add command -label $msg -state disabled
	    return
	}
	
	set cur_step [ GiD_Info postprocess get cur_step $cur_analysis]
	if { $cur_step == ""} {
	    $w delete 0 end
	    $w add command -label [_ "Please select a STEP first"] -state disabled
	    return
	}
	
	set all_res [ lsort -dictionary [ GiD_Info postprocess get cur_results_list $que_res]]
	if { $all_res == ""} {
	    $w delete 0 end
	    $w add command -label [_ "No useful results found"] -state disabled
	    return
	}
	
	if { $que_res == "Integrate_Vector_Normal" || $que_res == "Integrate_Vector_Tangential" } {
	    set also_components 0
	} else {
	    set also_components 1
	}
	
	## miramos si el menu que ya tenemos aun es valido
	# if { ![esMac]}  # en mac solo nos llaman cuando hay que rehacer los menus
	if { 1} { # en mac solo nos llaman cuando hay que rehacer los menus
	    # para el conyazo de windows, que es mas lento que una tortuga con reuma
	    if { ( $cur_analysis != "") && ( $cur_step != "") && ( $all_res != "") } {
		if { [ info exists VelasscoViewMenuPriv($w,$que_res//analstep)] && \
			 [ info exists VelasscoViewMenuPriv($w,$que_res//allres)] } {
		    if { ![ string compare $VelasscoViewMenuPriv($w,$que_res//analstep) "$cur_analysis $cur_step" ] || \
			     ![ string compare $VelasscoViewMenuPriv($w,$que_res//allres) $all_res] } {
			# ahora a mirar los componentes
			set igual 1
			foreach i $all_res {
			    if { $also_components } {
				set all_comp [ GiD_Info postprocess get cur_components_list $i ]
			    } else {
				set all_comp ""
			    }
			    regsub -all { } $i {_} idx
			    if { [ info exists VelasscoViewMenuPriv($w,$que_res//$idx)] } {
				if { [ string compare $VelasscoViewMenuPriv($w,$que_res//$idx) $all_comp]} {
				    # no son iguales
				    set igual 0
				    break
				}
			    } else {
				# no son iguales
				set igual 0
				break
			    }
			}
			if { $igual} {
			    return
			}
		    }
		}
	    }
	}
	
	# para ahorrarnos construir cosas
	set VelasscoViewMenuPriv($w,$que_res//analstep) "$cur_analysis $cur_step"
	set VelasscoViewMenuPriv($w,$que_res//allres) $all_res
	foreach i $all_res {
	    if { $also_components } {
		set all_comp [ GiD_Info postprocess get cur_components_list $i ]
	    } else {
		set all_comp ""
	    }
	    regsub -all { } $i {_} idx
	    set VelasscoViewMenuPriv($w,$que_res//$idx) $all_comp
	}
	
	$w delete 0 end
	
	if { $que_res == "Line_Graph"} {
	    set eje X
	    # construyamos el menu del eje X
	    if { [ winfo exists $w.m$eje ]} {
		$w.m$eje delete 0 end
	    } else {
		menu $w.m$eje
	    }
	    set cmds {X_Variation Y_Variation Z_Variation LineVariation}
	    set labels [ list [_ "X Variation"] [_ "Y Variation"] [_ "Z Variation"] [_ "Line Variation"]]
	    set values [ list "X_Variation" "Y_Variation" "Z_Variation" "LineVariation"]
	    set UnderLineList {}
	    if { ![ info exists ::VelasscoViewMenuPriv(LineGraphXAxis) ]} {
		set ::VelasscoViewMenuPriv(LineGraphXAxis) "LineVariation"
	    }
	    foreach i $cmds l $labels v $values {
		set under [FindUnderChar l UnderLineList]
		$w.m$eje add checkbutton -label $l -underline $under \
		    -onvalue $v -variable ::VelasscoViewMenuPriv(LineGraphXAxis)
	    }
	    $w add cascade -label [_ "Set X axis"] -menu $w.m$eje
	}
	
	set entrada_ventana 0
	if { $all_res != ""} {
	    foreach i $all_res {
		if { $also_components } {
		    if { $que_res != "Point_Complex_Evolution"} {
			set all_comp [ GiD_Info postprocess get cur_components_list $i ]
		    } else {
			set all_comp [ GiD_Info postprocess get cur_complex_components_list $i ]
		    }
		} else {
		    set all_comp ""
		}
		
		set wm [ GetResultMenuFromPath $w $i ""]
		if { $wm == ""} {
		    set entrada_ventana 1
		    break
		}
		
		set txt $i
		if { $wm != $w} {
		    if { [regexp {^/?/?(.*)//(.*?)$} $i dum path txt_res] } {
			set txt $txt_res
		    }                 
		}
		set txt_res [TranslateResultName $txt]
		
		set create_submenu 0
		if { [ llength $all_comp] > 1 } {
		    set create_submenu 1
		}
		
		# check if althought it's a scalar there are some created components elsewhere
		if { [ llength $all_comp] == 1} {
		    set root_res ${i}//
		    if { [ string first $root_res $all_res] != -1} {
			# may be of the form t//something
			# if { [ lsearch $all_res $root_res] != -1} 
			set create_submenu 1
			set txt_res [TranslateResultName $txt]
		    }
		    # WarnWinText "looking for '$root_res' in '$all_res' = [ string first $root_res $all_res]"
		}
		
		if { $create_submenu } {
		    # add component's menu
		    regsub -all {\.} $i {_} menu_name
		    $wm add cascade -label $txt_res -menu $wm.m$menu_name
		    if { [ winfo exists $wm.m$menu_name ]} {
			$wm.m$menu_name delete 0 end
		    } else {
			menu $wm.m$menu_name
		    }
		    foreach j $all_comp {
			set txt_comp [TranslateResultName $j]                   
			set do_graph_cmd [list {*}$basecommand $i $j]                    
			if { $que_res == "Line_Graph"} {
			    # set do_graph_cmd [list DoLineGraph $w {*}$basecommand $i $j]
			    set do_graph_cmd [list DoLineGraph $w $basecommand $i $j]
			}
			$wm.m$menu_name add command -label $txt_comp -command $do_graph_cmd
		    }
		} else {
		    # add the command
		    set do_graph_cmd [list {*}$basecommand $i]
		    if { $que_res == "Line_Graph"} {
			# set do_graph_cmd [list DoLineGraph $w {*}$basecommand $i]
			set do_graph_cmd [list DoLineGraph $w $basecommand $i]
		    }
		    $wm add command -label $txt_res -command $do_graph_cmd
		}
	    }
	}
	
	if { $entrada_ventana} {
	    $w add separator
	    $w add command -label ...[_ "More"]... -command PostGraphs::Create
	}
    }



    proc fromGiDEncoding { name} {
	regsub -all {_} $name { } ret_name
	return $ret_name
    }

    proc toGiDEncoding { name} {
	regsub -all { } $name {_} ret_name
	return $ret_name
    }

    proc isBusy { who} {
	variable _busy
	set busy 0
	if { [ info exists _busy($who)] && $_busy($who)} {
	    set bussy 1
	}
	return $busy
    }
    proc setBusy { who value} {
	variable _busy
	set _busy($who) $value
    }

    proc managePostprocessUpdateEvent { type args} {
	variable _cache 
	variable _selected

	# do my update:
	# get the results and do a contour fill
	# W "type = $type"
	# W "args = $args"

	switch $type {
	    "ResultViewEvent" {
		set res_view [ regsub -all {_} [ lindex $args 0] {}]
		# eventually it can be done for any result visualization ...
		# if { $res_view == "Contour_Fill"} {
		    set gid_res_name [ lindex $args 1]
		    set gid_comp_name [ lindex $args 2]
		    set res_name [ ::VELaSSCo::fromGiDEncoding $gid_res_name]
		    set comp_name [ ::VELaSSCo::fromGiDEncoding $gid_comp_name]
		    
		    # is result cached ?
		    # if not retrieve the results from VELaSSCo
		    # pass them to gid 
		    # and redo the contour fill
		    if { ![ AlreadyInGiD $_selected(analysis) $_selected(step) "$res_name"]} {
			set result_ids_values [ ::VELaSSCo::GetResultForVertices $res_name]
			set lst_ids [ lindex $result_ids_values 0]
			set lst_values [ lindex $result_ids_values 1]
			if { ( [ llength $lst_ids] != 0) && ( [ llength $lst_values] != 0) } {
			    CreateResultInGiD $_selected(analysis) $_selected(step) "$res_name" $lst_ids $lst_values
			    
			    #redo the contour fill after finishing the event
			    # GiD_Process Mescape Results ContourFill $gid_res_name $gid_comp_name
			    GiD_Process Mescape Results $res_view $gid_res_name $gid_comp_name
			} else {
			    WarnWin "Error retrieving the $res_name values from VELaSSCo: $result_ids_values"
			}
		    }
		# }
	    }
	    "LoadSelectionEvent" {
		set _selected(analysis) [ fromGiDEncoding [ lindex $args 0]]
		set _selected(step)  [ fromGiDEncoding [ lindex $args 1]]
	    }
	}

	if { [ llength [ info commands ::hidden_GiDPostprocessUpdate]] != 0} {
	    ::hidden_GiDPostprocessUpdate $type $args
	}
    }

    proc modifyGiDmenus { } {
	set idx [ GiDMenu::GetOptionIndex Files [ list "Recent post files"] POST]
	if { $idx == -1 } {
	    set idx end
	}
	# set icon [ ::VELaSSCo::GetImage Velassco16.png]
	set icon [ ::VELaSSCo::GetFullImageName Velassco16.png]
	# skip the --- separator just after Recent
	incr idx
	GiDMenu::InsertOption Files [ list "Connect to VELaSSCo ..."] $idx POST ::VELaSSCo::Login \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list "Open VELaSSCo model ..."] $idx POST ::VELaSSCo::SelectModel \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list ---] $idx "POST" "" "" "" "insertafter"
    }

    proc velasscoIconBarProc { { what DEFAULT} { img_dir ""}} {
	variable _dir
	variable _icon_do_connection
	variable _icon_connected
	global velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands velasscoIconBarBitmapsHelp
	
	set velasscoIconBarBitmapsNames(0) [ list Velassco24.png --- \
						 $_icon_do_connection --- \
						 open2.gif info.png \
						 volume.png boundary2d.png \
						 PostBarGraphPointEvolution.png PostBarGraphClear.png \
						 --- \
						 status.png VelasscoStop.png]
	
	set velasscoIconBarBitmapsCommands(0) [ list {-np- ::VELaSSCo::About} "" \
						    {-np- ::VELaSSCo::Login %W} "" \
						    {-np- ::VELaSSCo::SelectModel} {-np- ::VELaSSCo::GetModelInfo 1} \
						    {-np- ::VELaSSCo::DoGetBoundingBox} {-np- ::VELaSSCo::DoGetBoundaryOfAMesh} \
						    {-npm- ::VELaSSCo::AddGraphsResultsToMenu %W ::VELaSSCo::startNodeEvolution Point_Evolution} \
						    {Results Graphs OptionsGraph ClearGraphs} \
						    "" \
						    {-np- ::VELaSSCo::DoStatusDB} {-np- ::VELaSSCo::DoStopVELaSSCo} ]
	set velasscoIconBarBitmapsHelp(0) [ list "About VELaSSCo" "" \
						"Connect to VELaSSco server" "" \
						"Open a Model" "Show Model information" \
						"GetBoundingBox" "GetBoundaryOfAMesh" \
						"Node Evolution" \
						"" \
						"VELaSSCo status" "Stops QueryManager and StorageModule servers" ]

	# to create an empty pull-down menu for the node evolution graphs
	set velasscoIconBarBitmapsNames(0,8)	""
	set velasscoIconBarBitmapsCommands(0,8) ""

	CreateOtherBitmaps velasscoIconBar "VELaSSCo icon bar" velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands \
	    velasscoIconBarBitmapsHelp [ file join $_dir images] velasscoIconBarBitmaps $what Post
	
	AddNewToolbar velasscoIcon_bar PostVelasscoIconBarWindowGeom ::VELaSSCo::velasscoIconBarProc "VELaSSCo icon bar"
    }

}

proc register_myPostprocessUpdate {} {
    if { [ llength [ info commands ::hidden_GiDPostprocessUpdate]] != 0} {
	# already registered, nothing to do
	return
    }
    rename ::GiDPostprocessUpdate ::hidden_GiDPostprocessUpdate
    rename ::myPostprocessUpdate ::GiDPostprocessUpdate
}
proc unregister_myPostprocessUpdate {} {
    if { [ llength [ info commands ::VELaSSCo::myPostprocessUpdate]] != 0} {
	# already unregistered, nothing to do
	return
    }
    rename ::GiDPostprocessUpdate ::myPostprocessUpdate 
	rename ::hidden_GiDPostprocessUpdate ::GiDPostprocessUpdate 
}

proc InitGIDProject { dir} {
    ::VELaSSCo::Ini $dir

    # modify menus
    ::VELaSSCo::modifyGiDmenus

    # change to postprocess
    GiD_Process Mescape Postprocess

    # Create a new toolbar
    ::VELaSSCo::velasscoIconBarProc INSIDELEFT
    after 1000 ::VELaSSCo::velasscoIconBarProc INSIDELEFT
    after 1000 register_myPostprocessUpdate
}

proc EndGIDProject { } {
    unregister_myPostprocessUpdate
    ::VELaSSCo::End
}
proc EndGIDPostProcess { } {
    unregister_myPostprocessUpdate
    ::VELaSSCo::End
}
proc TclCalcModelBoundaries { useof} {
    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    ::VELaSSCo::TclCalcModelBoundaries $useof
}

proc myPostprocessUpdate { type args} {
    #program the action
    if { [ ::VELaSSCo::isBusy myPostprocessUpdate]} {
	return
    }
    ::VELaSSCo::setBusy myPostprocessUpdate 1
    after 500 [ list ::VELaSSCo::managePostprocessUpdateEvent $type {*}$args \n ::VELaSSCo::setBusy myPostprocessUpdate 0]
}
